@@ -3,26 +3,27 @@
 #include <string.h>
 #include <math.h>
 
+#define FUNC_isqrt 1
 #define FUNC_next_prime_in_sieve
 #include "sieve.h"
 #include "ptypes.h"
 #include "cache.h"
-#define FUNC_isqrt 1
 #include "util.h"
 #include "primality.h"
 #include "montmath.h"
+#include "prime_nth_count.h"
 
 /* Is it better to do a partial sieve + primality tests vs. full sieve? */
 static int do_partial_sieve(UV startp, UV endp) {
   UV range = endp - startp;
   if (USE_MONTMATH) range /= 8;  /* Fast primality tests */
 #if BITS_PER_WORD == 64
-  if ( (startp > UVCONST(     100000000000000) && range <     20000) ||
-       (startp > UVCONST(    1000000000000000) && range <    100000) ||
-       (startp > UVCONST(   10000000000000000) && range <    200000) ||
-       (startp > UVCONST(  100000000000000000) && range <   2000000) ||
+  if ( (startp > UVCONST(     100000000000000) && range <     40000) ||
+       (startp > UVCONST(    1000000000000000) && range <    150000) ||
+       (startp > UVCONST(   10000000000000000) && range <    600000) ||
+       (startp > UVCONST(  100000000000000000) && range <   2500000) ||
        (startp > UVCONST( 1000000000000000000) && range <  10000000) ||
-       (startp > UVCONST(10000000000000000000) && range <  20000000) )
+       (startp > UVCONST(10000000000000000000) && range <  40000000) )
     return 1;
 #endif
   return 0;
@@ -115,103 +116,6 @@ static const unsigned char presieve13[PRESIEVE_SIZE] =
   0x18,0x89,0x08,0x25,0x44,0x22,0x30,0x14,0xc3,0x88,0x86,0x40,0x1a,
   0x28,0x30,0x85,0x09,0x54,0x60,0x43,0x24,0x92,0x81,0x08,0x04,0x70};
 
-static const unsigned char stepdata[8][8][8] = {
-  { {96,65,34,67,36,69,102,47},
-    {64,47,75,106,46,105,77,44},
-    {32,86,33,87,115,45,114,84},
-    {72,46,85,42,73,127,44,123},
-    {96,35,100,79,33,66,37,78},
-    {64,108,34,109,67,47,65,46},
-    {32,76,117,33,118,74,35,87},
-    {32,127,86,45,84,43,82,121},
-  },
-  { {65,34,67,36,69,102,47,96},
-    {105,77,44,64,47,75,106,46},
-    {33,87,115,45,114,84,32,86},
-    {73,127,44,123,72,46,85,42},
-    {33,66,37,78,96,35,100,79},
-    {65,46,64,108,34,109,67,47},
-    {33,118,74,35,87,32,76,117},
-    {121,32,127,86,45,84,43,82},
-  },
-  { {34,67,36,69,102,47,96,65},
-    {106,46,105,77,44,64,47,75},
-    {114,84,32,86,33,87,115,45},
-    {42,73,127,44,123,72,46,85},
-    {66,37,78,96,35,100,79,33},
-    {34,109,67,47,65,46,64,108},
-    {74,35,87,32,76,117,33,118},
-    {82,121,32,127,86,45,84,43},
-  },
-  { {67,36,69,102,47,96,65,34},
-    {75,106,46,105,77,44,64,47},
-    {115,45,114,84,32,86,33,87},
-    {123,72,46,85,42,73,127,44},
-    {35,100,79,33,66,37,78,96},
-    {67,47,65,46,64,108,34,109},
-    {35,87,32,76,117,33,118,74},
-    {43,82,121,32,127,86,45,84},
-  },
-  { {36,69,102,47,96,65,34,67},
-    {44,64,47,75,106,46,105,77},
-    {84,32,86,33,87,115,45,114},
-    {44,123,72,46,85,42,73,127},
-    {100,79,33,66,37,78,96,35},
-    {108,34,109,67,47,65,46,64},
-    {76,117,33,118,74,35,87,32},
-    {84,43,82,121,32,127,86,45},
-  },
-  { {69,102,47,96,65,34,67,36},
-    {77,44,64,47,75,106,46,105},
-    {45,114,84,32,86,33,87,115},
-    {85,42,73,127,44,123,72,46},
-    {37,78,96,35,100,79,33,66},
-    {109,67,47,65,46,64,108,34},
-    {117,33,118,74,35,87,32,76},
-    {45,84,43,82,121,32,127,86},
-  },
-  { {102,47,96,65,34,67,36,69},
-    {46,105,77,44,64,47,75,106},
-    {86,33,87,115,45,114,84,32},
-    {46,85,42,73,127,44,123,72},
-    {78,96,35,100,79,33,66,37},
-    {46,64,108,34,109,67,47,65},
-    {118,74,35,87,32,76,117,33},
-    {86,45,84,43,82,121,32,127},
-  },
-  { {47,96,65,34,67,36,69,102},
-    {47,75,106,46,105,77,44,64},
-    {87,115,45,114,84,32,86,33},
-    {127,44,123,72,46,85,42,73},
-    {79,33,66,37,78,96,35,100},
-    {47,65,46,64,108,34,109,67},
-    {87,32,76,117,33,118,74,35},
-    {127,86,45,84,43,82,121,32},
-  },
-};
-
-static const int wheelmap[30] =
-  {0,0,0,0,0,0,0,1,0,0,0,2,0,3,0,0,0,4,0,5,0,0,0,6,0,0,0,0,0,7};
-static const int wheel2xmap[30] =     /* (2*p)%30 => 2,14,22,26,4,8,16,28 */
-  {0,0,0,0,4,0,0,0,5,0,0,0,0,0,1,0,6,0,0,0,0,0,2,0,0,0,3,0,7,0};
-/*     2   4       8          14  16          22      26  28     (2*p)%30 */
-
-#define FIND_COMPOSITE_POSITIONS(d, m, p) \
-  do { \
-    int v; \
-    UV dinc = (2*p) / 30; \
-    UV minc = (2*p) - dinc*30; \
-    const unsigned char* steps = stepdata [wheelmap[m]] [wheel2xmap[minc]]; \
-    v = steps[0]; wdinc[0] = dinc*(v>>5)+((v>>3)&0x3); wmask[0] = 1<<(v&0x7); \
-    v = steps[1]; wdinc[1] = dinc*(v>>5)+((v>>3)&0x3); wmask[1] = 1<<(v&0x7); \
-    v = steps[2]; wdinc[2] = dinc*(v>>5)+((v>>3)&0x3); wmask[2] = 1<<(v&0x7); \
-    v = steps[3]; wdinc[3] = dinc*(v>>5)+((v>>3)&0x3); wmask[3] = 1<<(v&0x7); \
-    v = steps[4]; wdinc[4] = dinc*(v>>5)+((v>>3)&0x3); wmask[4] = 1<<(v&0x7); \
-    v = steps[5]; wdinc[5] = dinc*(v>>5)+((v>>3)&0x3); wmask[5] = 1<<(v&0x7); \
-    v = steps[6]; wdinc[6] = dinc*(v>>5)+((v>>3)&0x3); wmask[6] = 1<<(v&0x7); \
-    v = steps[7]; wdinc[7] = dinc*(v>>5)+((v>>3)&0x3); wmask[7] = 1<<(v&0x7); \
-  } while (0)
-
 static const UV max_sieve_prime = (BITS_PER_WORD==64) ? 4294967291U : 65521U;
 
 /* Tile <from> bytes from source to <to> bytes in dest */
@@ -250,10 +154,94 @@ static UV sieve_prefill(unsigned char* mem, UV startd, UV endd)
   return vnext_prime;
 }
 
-/* Wheel 30 sieve.  Ideas from Terje Mathisen and Quesada / Van Pelt. */
+/* Marking primes is done the same way we used to do with tables, but
+ * now uses heavily unrolled code based on Kim Walisch's mod-30 sieve.
+ */
+#define set_bit(s,n)  *(s) |= (1 << n);
+static const unsigned char masknum30[30] =
+    {0,0,0,0,0,0,0,1,0,0,0,2,0,3,0,0,0,4,0,5,0,0,0,6,0,0,0,0,0,7};
+static const unsigned char qinit30[30] =
+    {0,0,1,1,1,1,1,1,2,2,2,2,3,3,4,4,4,4,5,5,6,6,6,6,7,7,7,7,7,7};
+
+typedef struct {
+  uint32_t prime;
+  uint32_t offset;
+  uint8_t index;
+} wheel_t;
+
+#define CROSS_INDEX(v, b0,b1,b2,b3,b4,b5,b6,b7,  i0,i1,i2,i3,i4,i5,i6,i7, it) \
+  while (1) { \
+    case (v+0): if(s>=send){w->index=v+0;break;} set_bit(s,b0); s += r*6+i0; \
+    case (v+1): if(s>=send){w->index=v+1;break;} set_bit(s,b1); s += r*4+i1; \
+    case (v+2): if(s>=send){w->index=v+2;break;} set_bit(s,b2); s += r*2+i2; \
+    case (v+3): if(s>=send){w->index=v+3;break;} set_bit(s,b3); s += r*4+i3; \
+    case (v+4): if(s>=send){w->index=v+4;break;} set_bit(s,b4); s += r*2+i4; \
+    case (v+5): if(s>=send){w->index=v+5;break;} set_bit(s,b5); s += r*4+i5; \
+    case (v+6): if(s>=send){w->index=v+6;break;} set_bit(s,b6); s += r*6+i6; \
+    case (v+7): if(s>=send){w->index=v+7;break;} set_bit(s,b7); s += r*2+i7; \
+    while (s + r*28 + it-1 < send) { \
+      set_bit(s + r *  0 +  0, b0); \
+      set_bit(s + r *  6 + i0, b1); \
+      set_bit(s + r * 10 + i0+i1, b2); \
+      set_bit(s + r * 12 + i0+i1+i2, b3); \
+      set_bit(s + r * 16 + i0+i1+i2+i3, b4); \
+      set_bit(s + r * 18 + i0+i1+i2+i3+i4, b5); \
+      set_bit(s + r * 22 + i0+i1+i2+i3+i4+i5, b6); \
+      set_bit(s + r * 28 + i0+i1+i2+i3+i4+i5+i6, b7); \
+      s += r*30 + it; \
+    } \
+  }
+static wheel_t create_wheel(UV startp, uint32_t prime)
+{
+  wheel_t w;
+  UV q = prime;
+  UV p2 = q*q;
+
+  if (startp == 0) {
+    wheel_t ws = { prime, p2/30, qinit30[q % 30]  +  8*masknum30[prime % 30] };
+    return ws;
+  }
+
+  if (p2 < startp) {
+    q = 1+(startp-1)/prime;
+    q += distancewheel30[q % 30];
+    p2 = prime * q;
+    /* The offset if p2 overflows is still ok, or set to max_sieve_prime+1. */
+    /* if (p2 < startp) p2 = max_sieve_prime+1; */
+  }
+
+  w.offset = (p2-startp) / 30;
+  w.index = qinit30[q % 30]  +  8*masknum30[prime % 30];
+  w.prime = prime;
+  return w;
+}
+
+static void mark_primes(unsigned char* s, uint32_t bytes, wheel_t* w)
+{
+  if (w->offset >= bytes) {
+    w->offset -= bytes;
+  } else {
+    const unsigned char* send = s + bytes;
+    uint32_t r = w->prime / 30;
+    s += w->offset;
+    switch (w->index) {
+      CROSS_INDEX( 0, 0,1,2,3,4,5,6,7, 0,0,0,0,0,0,0,1,  1); break;
+      CROSS_INDEX( 8, 1,5,4,0,7,3,2,6, 1,1,1,0,1,1,1,1,  7); break;
+      CROSS_INDEX(16, 2,4,0,6,1,7,3,5, 2,2,0,2,0,2,2,1, 11); break;
+      CROSS_INDEX(24, 3,0,6,5,2,1,7,4, 3,1,1,2,1,1,3,1, 13); break;
+      CROSS_INDEX(32, 4,7,1,2,5,6,0,3, 3,3,1,2,1,3,3,1, 17); break;
+      CROSS_INDEX(40, 5,3,7,1,6,0,4,2, 4,2,2,2,2,2,4,1, 19); break;
+      CROSS_INDEX(48, 6,2,3,7,0,4,5,1, 5,3,1,4,1,3,5,1, 23); break;
+      CROSS_INDEX(56, 7,6,5,4,3,2,1,0, 6,4,2,4,2,4,6,1, 29); break;
+    }
+    w->offset = s - send;
+  }
+}
+
+/* Monolithic mod-30 wheel sieve */
 unsigned char* sieve_erat30(UV end)
 {
-  unsigned char* mem;
+  unsigned char *mem;
   UV max_buf, limit, prime;
 
   max_buf = (end/30) + ((end%30) != 0);
@@ -266,56 +254,19 @@ unsigned char* sieve_erat30(UV end)
 
   limit = isqrt(end);  /* prime*prime can overflow */
   for (  ; prime <= limit; prime = next_prime_in_sieve(mem,prime,end)) {
-    UV p2 = prime*prime;
-    UV d = p2 / 30;
-    UV m = p2 - d*30;
-    UV wdinc[8];
-    unsigned char wmask[8];
-
-    /* Find the positions of the next composites we will mark */
-    FIND_COMPOSITE_POSITIONS(d, m, prime);
-#if 0
-    assert(d == ((prime*prime)/30));
-    assert(d < max_buf);
-    assert(prime = (wdinc[0]+wdinc[1]+wdinc[2]+wdinc[3]+wdinc[4]+wdinc[5]+wdinc[6]+wdinc[7]));
-#endif
-    /* Regular code to mark composites:
-    *  i = 0;
-    *  do {mem[d] |= wmask[i]; d += wdinc[i]; i = (i+1)&7;} while (d < max_buf);
-    * Unrolled version: */
-    while ( (d+prime) < max_buf ) {
-      mem[d] |= wmask[0];  d += wdinc[0];
-      mem[d] |= wmask[1];  d += wdinc[1];
-      mem[d] |= wmask[2];  d += wdinc[2];
-      mem[d] |= wmask[3];  d += wdinc[3];
-      mem[d] |= wmask[4];  d += wdinc[4];
-      mem[d] |= wmask[5];  d += wdinc[5];
-      mem[d] |= wmask[6];  d += wdinc[6];
-      mem[d] |= wmask[7];  d += wdinc[7];
-    }
-    while (1) {
-      mem[d] |= wmask[0];  d += wdinc[0];  if (d >= max_buf) break;
-      mem[d] |= wmask[1];  d += wdinc[1];  if (d >= max_buf) break;
-      mem[d] |= wmask[2];  d += wdinc[2];  if (d >= max_buf) break;
-      mem[d] |= wmask[3];  d += wdinc[3];  if (d >= max_buf) break;
-      mem[d] |= wmask[4];  d += wdinc[4];  if (d >= max_buf) break;
-      mem[d] |= wmask[5];  d += wdinc[5];  if (d >= max_buf) break;
-      mem[d] |= wmask[6];  d += wdinc[6];  if (d >= max_buf) break;
-      mem[d] |= wmask[7];  d += wdinc[7];  if (d >= max_buf) break;
-    }
+    wheel_t w = create_wheel(0, prime);
+    mark_primes(mem, max_buf, &w);
   }
   return mem;
 }
 
-
-
+/* Segmented mod-30 wheel sieve */
 int sieve_segment(unsigned char* mem, UV startd, UV endd)
 {
   const unsigned char* sieve;
   UV limit, slimit, start_base_prime, sieve_size;
   UV startp = 30*startd;
   UV endp = (endd >= (UV_MAX/30))  ?  UV_MAX-2  :  30*endd+29;
-
   MPUassert( (mem != 0) && (endd >= startd) && (endp >= startp),
              "sieve_segment bad arguments");
 
@@ -330,8 +281,7 @@ int sieve_segment(unsigned char* mem, UV startd, UV endd)
   /* Fill buffer with marked 7, 11, and 13 */
   start_base_prime = sieve_prefill(mem, startd, endd);
 
-  limit = isqrt(endp);  /* floor(sqrt(n)), will include p if p*p=endp */
-  /* Don't use a sieve prime such that p*p > UV_MAX */
+  limit = isqrt(endp);
   if (limit > max_sieve_prime)  limit = max_sieve_prime;
   slimit = limit;
   if (do_partial_sieve(startp, endp))
@@ -342,78 +292,56 @@ int sieve_segment(unsigned char* mem, UV startd, UV endd)
     get_prime_cache(slimit, &sieve);
   }
 
-  START_DO_FOR_EACH_SIEVE_PRIME(sieve, 0, start_base_prime, slimit)
-  {
-    /* p increments from 17 to at most sqrt(endp).  Note on overflow:
-     * 32-bit: limit=     65535, max p =      65521, p*p = ~0-1965854
-     * 64-bit: limit=4294967295, max p = 4294967291, p*p = ~0-42949672934
-     * No overflow here, but possible after the incrementing below. */
-    UV p2 = p*p;
-    if (p2 < startp) {
-      UV f = 1+(startp-1)/p;
-      p2 = p * (f + distancewheel30[f%30]);
-    }
-    /* It is possible we've overflowed p2, so check for that */
-    if ( (p2 <= endp) && (p2 >= startp) ) {
-      /* Sieve from startd to endd starting at p2, stepping p */
-#if 0
-      /* Basic sieve */
-      do {
-        mem[(p2 - startp)/30] |= masktab30[p2%30];
-        do { p2 += 2*p; } while (masktab30[p2%30] == 0);
-      } while ( (p2 <= endp) && (p2 >= startp) );
-#else
-      UV d = p2 / 30;
-      UV m = p2 - d*30;
-
-      if ((p2 + 2*p) > endp) {
-        /* There is only one composite to be marked in this segment */
-        mem[d-startd] |= masktab30[m];
-      } else {
-        UV wdinc[8];
-        unsigned char wmask[8];
-        UV offset_endd = endd - startd;
-        UV unrolls = (endd-d+1) / p;
-        /* Find the positions of the next composites we will mark */
-        FIND_COMPOSITE_POSITIONS(d, m, p);
-        d -= startd;
-        /* Unrolled inner loop for marking composites */
-        while ( unrolls-- > 0) {
-          mem[d] |= wmask[0];  d += wdinc[0];
-          mem[d] |= wmask[1];  d += wdinc[1];
-          mem[d] |= wmask[2];  d += wdinc[2];
-          mem[d] |= wmask[3];  d += wdinc[3];
-          mem[d] |= wmask[4];  d += wdinc[4];
-          mem[d] |= wmask[5];  d += wdinc[5];
-          mem[d] |= wmask[6];  d += wdinc[6];
-          mem[d] |= wmask[7];  d += wdinc[7];
-        }
-        while (d <= offset_endd) {
-          mem[d] |= wmask[0];  d += wdinc[0];  if (d > offset_endd) break;
-          mem[d] |= wmask[1];  d += wdinc[1];  if (d > offset_endd) break;
-          mem[d] |= wmask[2];  d += wdinc[2];  if (d > offset_endd) break;
-          mem[d] |= wmask[3];  d += wdinc[3];  if (d > offset_endd) break;
-          mem[d] |= wmask[4];  d += wdinc[4];  if (d > offset_endd) break;
-          mem[d] |= wmask[5];  d += wdinc[5];  if (d > offset_endd) break;
-          mem[d] |= wmask[6];  d += wdinc[6];  if (d > offset_endd) break;
-          mem[d] |= wmask[7];  d += wdinc[7];
-        }
-      }
-#endif
-    }
-  }
-  END_DO_FOR_EACH_SIEVE_PRIME;
+  START_DO_FOR_EACH_SIEVE_PRIME(sieve, 0, start_base_prime, slimit) {
+    wheel_t w = create_wheel(startp, p);
+    mark_primes(mem, endd-startd+1, &w);
+  } END_DO_FOR_EACH_SIEVE_PRIME;
   release_prime_cache(sieve);
 
   if (limit > slimit) { /* We've sieved out most composites, but not all. */
     START_DO_FOR_EACH_SIEVE_PRIME(mem, 0, 0, endp-startp) {
-      if (!BPSW(startp + p))        /* If the candidate is not prime, */
-        mem[d_] |= 1 << bit_;       /* mark the sieve location.       */
+      if (!BPSW(startp + p))           /* If the candidate is not prime, */
+        mem[p/30] |= masktab30[p%30];  /* mark the sieve location.       */
     } END_DO_FOR_EACH_SIEVE_PRIME;
   }
   return 1;
 }
 
+int sieve_segment_wheel(unsigned char* mem, UV startd, UV endd, wheel_t *warray, uint32_t wsize)
+{
+  uint32_t i = 0, limit, start_base_prime;
+  uint32_t segsize = endd - startd + 1;
+  UV startp = 30*startd;
+  UV endp = (endd >= (UV_MAX/30))  ?  UV_MAX-2  :  30*endd+29;
+  MPUassert( (mem != 0) && (endd >= startd) && (endp >= startp),
+             "sieve_segment bad arguments");
+
+  /* possibly use primary cache directly */
+
+  /* Fill buffer with marked 7, 11, and 13 */
+  start_base_prime = sieve_prefill(mem, startd, endd);
+  while (i < wsize && warray[i].prime < start_base_prime)
+    i++;
+
+  limit = isqrt(endp);
+  if (limit > max_sieve_prime)  limit = max_sieve_prime;
+
+  while (i < wsize && warray[i].prime <= limit) {
+    if (warray[i].index >= 64)
+      warray[i] = create_wheel(startp, warray[i].prime);
+    mark_primes(mem, segsize, &(warray[i++]));
+  }
+
+  if (limit > warray[wsize-1].prime && warray[wsize-1].prime < max_sieve_prime) {
+    START_DO_FOR_EACH_SIEVE_PRIME(mem, 0, 0, endp-startp) {
+      if (!BPSW(startp + p))           /* If the candidate is not prime, */
+        mem[p/30] |= masktab30[p%30];  /* mark the sieve location.       */
+    } END_DO_FOR_EACH_SIEVE_PRIME;
+  }
+
+  return 1;
+}
+
 /**************************************************************************/
 
 typedef struct {
@@ -425,6 +353,8 @@ typedef struct {
   UV segment_size;
   unsigned char* segment;
   unsigned char* base;
+  wheel_t *warray;
+  uint32_t wsize;
 } segment_context_t;
 
 /*
@@ -447,7 +377,7 @@ typedef struct {
 void* start_segment_primes(UV low, UV high, unsigned char** segmentmem)
 {
   segment_context_t* ctx;
-  UV slimit;
+  UV nsegments, range;
 
   MPUassert( high >= low, "start_segment_primes bad arguments");
   New(0, ctx, 1, segment_context_t);
@@ -456,29 +386,61 @@ void* start_segment_primes(UV low, UV high, unsigned char** segmentmem)
   ctx->lod = low / 30;
   ctx->hid = high / 30;
   ctx->endp = (ctx->hid >= (UV_MAX/30))  ?  UV_MAX-2  :  30*ctx->hid+29;
+  range = ctx->hid - ctx->lod + 1;   /* range in bytes */
 
 #if BITS_PER_WORD == 64
-  if (high > 1e11 && high-low > 1e6) {
-    UV range = (high-low+29)/30;
-    /* Select what we think would be a good segment size */
-    UV size = isqrt(isqrt(high)) * ((high < 1e15) ? 500 : 250);
+  if (high > 1e10 && range > 32*1024-16) {
+    UV size, div;
+    /* Use larger segments */
+    size = isqrt(32*isqrt(high)) * (logint(high,2)-2);
+    if (size < 128*1024) size = 128*1024;
     /* Evenly split the range into segments */
-    UV div = (range+size-1)/size;
+    div = (range+size-1)/size;
     size = (div <= 1)  ?  range  :  (range+div-1)/div;
-    if (_XS_get_verbose() >= 2)
-      printf("segment sieve: byte range %lu split into %lu segments of size %lu\n", (unsigned long)range, (unsigned long)div, (unsigned long)size);
     ctx->segment_size = size;
     New(0, ctx->segment, size, unsigned char);
   } else
 #endif
   ctx->segment = get_prime_segment( &(ctx->segment_size) );
   *segmentmem = ctx->segment;
+  nsegments = (((high-low+29)/30)+ctx->segment_size-1) / ctx->segment_size;
+
+  if (_XS_get_verbose() >= 2)
+    printf("segment sieve: byte range %lu split into %lu segments of size %lu\n", (unsigned long)range, (unsigned long)nsegments, (unsigned long)ctx->segment_size);
 
   ctx->base = 0;
-  /* Expand primary cache so we won't regen each call */
-  slimit = isqrt(ctx->endp)+1;
-  if (do_partial_sieve(low, high))  slimit >>= 8;
-  get_prime_cache( slimit, 0);
+  ctx->warray = 0;
+  ctx->wsize = 0;
+#if 1
+  { /* Generate wheel data for this segment sieve */
+    const UV maxsieve = UVCONST(400000000);
+    UV limit, nprimes;
+    wheel_t *warray;
+    wheel_t w = {0,0,128};
+    uint32_t wsize = 0;
+    /* Number of primes for a full sieve */
+    limit = isqrt(ctx->endp);
+    /* For small ranges a partial sieve is much faster */
+    if (do_partial_sieve(low, high))
+      limit >>= ((low < (UV)1e16) ? 8 : 10);
+    if (limit <= maxsieve) {
+      /* Bump to one more than needed. */
+      limit = next_prime(limit);
+      /* We'll make space for this many */
+      nprimes = prime_count_upper(limit);
+      if (_XS_get_verbose() >= 4)
+        printf("segment sieve %lu - %lu, primes to %lu (max %lu)\n", (unsigned long)low, (unsigned long)high, (unsigned long)limit, (unsigned long)nprimes);
+      New(0, warray, nprimes, wheel_t);
+      START_DO_FOR_EACH_PRIME(0,limit) {
+        if (wsize >= nprimes) croak("segment bad upper count");
+        w.prime = p;
+        warray[wsize++] = w;
+      } END_DO_FOR_EACH_PRIME;
+      ctx->warray = warray;
+      ctx->wsize = wsize;
+    }
+  }
+#endif
 
   return (void*) ctx;
 }
@@ -501,7 +463,10 @@ int next_segment_primes(void* vctx, UV* base, UV* low, UV* high)
   MPUassert( seghigh_d >= ctx->lod, "next_segment_primes: highd < lowd");
   MPUassert( range_d <= ctx->segment_size, "next_segment_primes: range > segment size");
 
-  sieve_segment(ctx->segment, ctx->lod, seghigh_d);
+  if (ctx->warray != 0)
+    sieve_segment_wheel(ctx->segment, ctx->lod, seghigh_d, ctx->warray, ctx->wsize);
+  else
+    sieve_segment(ctx->segment, ctx->lod, seghigh_d);
 
   ctx->lod += range_d;
   ctx->low = *high + 2;
@@ -521,5 +486,9 @@ void end_segment_primes(void* vctx)
     Safefree(ctx->base);
     ctx->base = 0;
   }
+  if (ctx->warray != 0) {
+    Safefree(ctx->warray);
+    ctx->warray = 0;
+  }
   Safefree(ctx);
 }
